@@ -22,35 +22,40 @@
 #  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 #  USA
 
+from __future__ import absolute_import, print_function
+
 import apt
 import apt_pkg
 import dbus
-import tempfile
 from gettext import gettext as _
+import gettext
 import os
+import subprocess
+from aptdaemon import client
+from aptdaemon.errors import NotAuthorizedError, TransactionFailed
 
 from gi.repository import GObject, Gdk, Gtk, Gio
 
-from SimpleGtkbuilderApp import SimpleGtkbuilderApp
-from aptsources.sourceslist import SourceEntry
-from DialogAdd import DialogAdd
-from DialogMirror import DialogMirror
-from DialogEdit import DialogEdit
-from DialogCacheOutdated import DialogCacheOutdated
-from DialogAddSourcesList import DialogAddSourcesList
-from CdromProgress import CdromProgress
+from .SimpleGtkbuilderApp import SimpleGtkbuilderApp
+from .DialogAdd import DialogAdd
+from .DialogMirror import DialogMirror
+from .DialogEdit import DialogEdit
+from .DialogCacheOutdated import DialogCacheOutdated
+from .DialogAddSourcesList import DialogAddSourcesList
 
 import softwareproperties
 import softwareproperties.distro
 from softwareproperties.SoftwareProperties import SoftwareProperties
 import softwareproperties.SoftwareProperties
 
+from UbuntuDrivers import detect
+
 (LIST_MARKUP, LIST_ENABLED, LIST_ENTRY_OBJ) = range(3)
 
 (
     COLUMN_ACTIVE,
     COLUMN_DESC
-) = range(2)
+) = list(range(2))
 
 RESPONSE_REPLACE = 1
 RESPONSE_ADD = 2
@@ -62,7 +67,7 @@ RESPONSE_ADD = 2
     STORE_SOURCE, 
     STORE_SEPARATOR,
     STORE_VISIBLE
-) = range(5)
+) = list(range(5))
 
 def error(parent_window, summary, msg):
     """ show a error dialog """
@@ -72,13 +77,15 @@ def error(parent_window, summary, msg):
                                buttons=Gtk.ButtonsType.OK,
                                message_format=None)
     dialog.set_markup("<big><b>%s</b></big>\n\n%s" % (summary, msg))
-    res = dialog.run()
+    dialog.run()
     dialog.destroy()
     return False
 
 # using this often, so defining it here.
 def permission_denied_dialog(parent_window):
-    print("Warning: Authorization failed. Software sources can't be changed without permission.")
+    error(parent_window,
+          _("Authorization failed."), 
+          _("Software sources can't be changed without permission."))
 
 class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
 
@@ -150,8 +157,10 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
     # used to store the handlers of callbacks
     self.handlers = []
 
+    self.apt_cache = apt.Cache()
+    self.apt_client = client.AptClient()
+
     # Put some life into the user interface:
-    self.init_popcon()
     self.init_auto_update()
     self.init_release_upgrades()
     self.show_auto_update_level()
@@ -165,7 +174,11 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
     # Setup and show the distro elements
     self.init_distro()
     self.show_distro()
+    # Setup and show the Additonal Drivers tab
+    self.init_drivers()
+    self.show_drivers()
 
+    self.notebook_main.connect("switch-page", self.on_main_notebook_page_switched)
     # Show the import/replace sources.list dialog if a file different
     # to the default sources.list was specified 
     # NOTE: If the file path points to the default sources.list the user
@@ -178,21 +191,18 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
                                            apt_pkg.config.find("Dir::Etc::sourcelist"))):
         self.open_file(file)
 
+  def on_main_notebook_page_switched(self, notebook, page, page_num):
+    # On the additional drivers page, don't show the backend revert button.
+    if page == self.vbox_drivers:
+      self.button_revert.set_visible(False)
+    else:
+      self.button_revert.set_visible(True)
+
   def update_interface(self):
     """ abstract interface to keep the UI alive """
     while Gtk.events_pending():
       Gtk.main_iteration()
 
-  def init_popcon(self):
-    """ If popcon is enabled show the statistics tab and an explanation
-        corresponding to the used distro """
-    is_helpful = self.get_popcon_participation()
-    if is_helpful != None:
-      self.label_popcon_desc.set_label(
-          softwareproperties.distro.get_popcon_description(self.distro))
-      self.vbox_popcon.show()
-      self.checkbutton_popcon.set_active(is_helpful)
-
   def init_release_upgrades(self):
     " setup the widgets that allow configuring the release upgrades "
     i = self.get_release_upgrades_policy()
@@ -234,15 +244,6 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
          self.combobox_update_interval.connect("changed", 
                                      self.on_combobox_update_interval_changed)))
     
-    self.handlers.append(
-        (self.combobox_security_updates,
-         self.combobox_security_updates.connect("changed",
-                                     self.set_sec_update_automation_level)))
-    
-    self.handlers.append(
-        (self.combobox_security_updates,
-         self.combobox_other_updates.connect("changed",
-                                     self.set_other_update_automation_level)))
   
   def show_auto_update_level(self):
     """Represent the level of update automation in the user interface"""
@@ -272,11 +273,21 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
                 self.combobox_other_updates.set_active(i)
                 break
 
+    self.handlers.append(
+        (self.combobox_security_updates,
+         self.combobox_security_updates.connect("changed",
+                                     self.set_sec_update_automation_level)))
+    
+    self.handlers.append(
+        (self.combobox_other_updates,
+         self.combobox_other_updates.connect("changed",
+                                     self.set_other_update_automation_level)))
+
   def init_distro(self):
     """Setup the user interface elements to represent the distro"""
 
     # TRANS: %s stands for the distribution name e.g. Debian or Ubuntu
-    self.label_dist_name.set_label(_("%s Software") % self.distro.id.encode('UTF-8'))
+    self.label_dist_name.set_label(_("%s Software") % self.distro.id)
 
     self.handlers.append((self.checkbutton_source_code,
                           self.checkbutton_source_code.connect("toggled",
@@ -415,24 +426,24 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
 
     # Output a lot of debug stuff
     if self.options.debug == True or self.options.massive_debug == True:
-        print "ENABLED COMPS: %s" % self.distro.enabled_comps
-        print "INTERNET COMPS: %s" % self.distro.download_comps
-        print "MAIN SOURCES"
+        print("ENABLED COMPS: %s" % self.distro.enabled_comps)
+        print("INTERNET COMPS: %s" % self.distro.download_comps)
+        print("MAIN SOURCES")
         for source in self.distro.main_sources:
             self.print_source_entry(source)
-        print "CHILD SOURCES"
+        print("CHILD SOURCES")
         for source in self.distro.child_sources:
             self.print_source_entry(source)
-        print "CDROM SOURCES"
+        print("CDROM SOURCES")
         for source in self.distro.cdrom_sources:
             self.print_source_entry(source)
-        print "SOURCE CODE SOURCES"
+        print("SOURCE CODE SOURCES")
         for source in self.distro.source_code_sources:
             self.print_source_entry(source)
-        print "DISABLED SOURCES"
+        print("DISABLED SOURCES")
         for source in self.distro.disabled_sources:
             self.print_source_entry(source)
-        print "ISV"
+        print("ISV")
         for source in self.sourceslist_visible:
             self.print_source_entry(source)
 
@@ -469,7 +480,7 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
 
   def on_combobox_release_upgrades_changed(self, combobox):
     """ set the release upgrades policy """
-    #print "on_combobox_release_upgrades_changed()"
+    #print("on_combobox_release_upgrades_changed()")
     i = combobox.get_active()
     try:
         self.backend.SetReleaseUpgradesPolicy(i)
@@ -589,6 +600,10 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
   def hide(self):
     self.window_main.hide()
     
+  def source_search_function(self, store, column, key, iter, data=None):
+    # this needs to return False on a match
+    return store.get_value(iter, COLUMN_DESC).lower().find(key.lower()) == -1
+
   def init_isv_sources(self):
     """
     Read all valid sources into our ListStore
@@ -652,6 +667,9 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
 
     self.treeview_sources.append_column(col_active)
     self.treeview_sources.append_column(col_desc)
+    # typeahead support for the sources list using case insensitive matching
+    self.treeview_sources.set_search_column(COLUMN_DESC)
+    self.treeview_sources.set_search_equal_func(self.source_search_function, None)
     # drag and drop support for sources.list
     try:
         Gtk.drag_dest_set(self.treeview_sources, Gtk.DestDefaults.ALL,
@@ -683,7 +701,7 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
     self.keys_store = Gtk.ListStore(str)
     self.treeview_auth.set_model(self.keys_store)
     tr = Gtk.CellRendererText()
-    keys_col = Gtk.TreeViewColumn("Key", tr, text=0)
+    keys_col = Gtk.TreeViewColumn(_("Key"), tr, text=0)
     self.treeview_auth.append_column(keys_col)
     try:
         self.treeview_auth.enable_model_drag_dest(
@@ -759,12 +777,10 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
     self.show_isv_sources()
     self.show_cdrom_sources()
     self.button_revert.set_sensitive(True)
-    self.modified_sourceslist = True
 
   def on_auth_failed(self):
     """ send when the authentication failed """
     # reread the current config
-    self.init_popcon()
     self.on_sources_list_modified()
     self.on_config_modified()
     self.on_keys_modified()
@@ -786,7 +802,7 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
                                   source, False, True])
 
     (path_x, path_y) = self.treeview_sources.get_cursor()
-    if len(self.source_store) < 1 or path_x <0:
+    if len(self.source_store) < 1 or path_x is None:
         self.button_remove.set_sensitive(False)
         self.button_edit.set_sensitive(False)
 
@@ -803,8 +819,8 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
     """ Return true if the selected row is a separator """
     try:
       return model.get_value(iter, column)
-    except Exception, e:
-      print "is_seperator returned '%s' " % e
+    except Exception as e:
+      print("is_seperator returned '%s' " % e)
       return False
       
   def show_keys(self):
@@ -953,13 +969,370 @@ class SoftwarePropertiesGtk(SoftwareProperties,SimpleGtkbuilderApp):
         self.options.no_update == False):
         d = DialogCacheOutdated(self.window_main,
                                 self.datadir)
-        res = d.run()
+        d.run()
     self.quit()
 
   def on_button_add_cdrom_clicked(self, widget):
       """ when a cdrom is requested for adding """
       try:
-          self.backend.AddCdromSource()
+        self.backend.AddCdromSource()
       except dbus.DBusException as e:
-          if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-              permission_denied_dialog(self.window_main)
+        if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
+            permission_denied_dialog(self.window_main)      
+
+  def on_driver_changes_progress(self, transaction, progress):
+    #print(progress)
+    self.button_driver_revert.set_visible(False)
+    self.button_driver_apply.set_visible(False)
+    self.button_driver_restart.set_visible(False)
+    self.button_driver_cancel.set_visible(True)
+    self.progress_bar.set_visible(True)
+    self.progress_bar.set_visible(True)
+
+    self.label_driver_action.set_label(_("Applying changes..."))
+    self.progress_bar.set_fraction(progress / 100.0)
+
+  def on_driver_changes_finish(self, transaction, exit_state):
+    self.progress_bar.set_visible(False)
+    self.clear_changes()
+    self.apt_cache = apt.Cache()
+    self.set_driver_action_status()
+    self.update_label_and_icons_from_status()
+    self.button_driver_revert.set_visible(True)
+    self.button_driver_apply.set_visible(True)
+    self.button_driver_cancel.set_visible(False)
+    self.scrolled_window_drivers.set_sensitive(True)
+
+  def on_driver_changes_error(self, transaction, error_code, error_details):
+    self.on_driver_changes_revert()
+    self.set_driver_action_status()
+    self.update_label_and_icons_from_status()
+    self.button_driver_revert.set_visible(True)
+    self.button_driver_apply.set_visible(True)
+    self.button_driver_cancel.set_visible(False)
+    self.scrolled_window_drivers.set_sensitive(True)
+
+  def on_driver_changes_cancellable_changed(self, transaction, cancellable):
+    self.button_driver_cancel.set_sensitive(cancellable)
+
+  def on_driver_changes_apply(self, button):
+
+    installs = []
+    removals = []
+
+    for pkg in self.driver_changes:
+      if pkg.is_installed:
+        removals.append(pkg.shortname)
+      else:
+        installs.append(pkg.shortname)
+
+    try:
+      self.transaction = self.apt_client.commit_packages(install=installs, remove=removals,
+                                                         reinstall=[], purge=[], upgrade=[], downgrade=[])
+      self.transaction.connect("progress-changed", self.on_driver_changes_progress)
+      self.transaction.connect("cancellable-changed", self.on_driver_changes_cancellable_changed)
+      self.transaction.connect("finished", self.on_driver_changes_finish)
+      self.transaction.connect("error", self.on_driver_changes_error)
+      self.transaction.run()
+      self.button_driver_revert.set_sensitive(False)
+      self.button_driver_apply.set_sensitive(False)
+      self.scrolled_window_drivers.set_sensitive(False)
+    except (NotAuthorizedError, TransactionFailed) as e:
+      print("Warning: install transaction not completed successfully: {}".format(e))
+
+
+  def on_driver_changes_revert(self, button_revert=None):
+
+    # HACK: set all the "Do not use" first; then go through the list of the
+    #       actually selected drivers.
+    for button in self.no_drv:
+      button.set_active(True)
+
+    for alias in self.orig_selection:
+      button = self.orig_selection[alias]
+      button.set_active(True)
+
+    self.clear_changes()
+
+    self.button_driver_revert.set_sensitive(False)
+    self.button_driver_apply.set_sensitive(False)
+
+  def on_driver_changes_cancel(self, button_cancel):
+    self.transaction.cancel()
+    self.clear_changes()
+
+  def on_driver_restart_clicked(self, button_restart):
+    subprocess.call(['/usr/lib/indicator-session/gtk-logout-helper', '--shutdown'])
+
+  def clear_changes(self):
+    self.orig_selection = {}
+    self.driver_changes = []
+
+  def init_drivers(self):
+    """Additional Drivers tab"""
+
+    self.button_driver_revert = Gtk.Button(label=_("Re_vert"), use_underline=True)
+    self.button_driver_revert.connect("clicked", self.on_driver_changes_revert)
+    self.button_driver_apply = Gtk.Button(label=_("_Apply Changes"), use_underline=True)
+    self.button_driver_apply.connect("clicked", self.on_driver_changes_apply)
+    self.button_driver_cancel = Gtk.Button(label=_("_Cancel"), use_underline=True)
+    self.button_driver_cancel.connect("clicked", self.on_driver_changes_cancel)
+    self.button_driver_restart = Gtk.Button(label=_("_Restart..."), use_underline=True)
+    self.button_driver_restart.connect("clicked", self.on_driver_restart_clicked)
+    self.button_driver_revert.set_sensitive(False)
+    self.button_driver_revert.set_visible(True)
+    self.button_driver_apply.set_sensitive(False)
+    self.button_driver_apply.set_visible(True)
+    self.button_driver_cancel.set_visible(False)
+    self.button_driver_restart.set_visible(False)
+    self.box_driver_action.pack_end(self.button_driver_apply, False, False, 0)
+    self.box_driver_action.pack_end(self.button_driver_revert, False, False, 0)
+    self.box_driver_action.pack_end(self.button_driver_restart, False, False, 0)
+    self.box_driver_action.pack_end(self.button_driver_cancel, False, False, 0)
+
+    self.progress_bar = Gtk.ProgressBar()
+    self.box_driver_action.pack_end(self.progress_bar, False, False, 0)
+    self.progress_bar.set_visible(False)
+
+    self.devices = detect.system_device_drivers()
+    self.driver_changes = []
+    self.orig_selection = {}
+    # HACK: the case where the selection is actually "Do not use"; is a little
+    #       tricky to implement because you can't check for whether a package is
+    #       installed or any such thing. So let's keep a list of all the 
+    #       "Do not use" radios, set those active first, then iterate through
+    #       orig_selection when doing a Reset.
+    self.no_drv = []
+    self.nonfree_drivers = 0
+    self.ui_building = False
+
+  def on_driver_selection_changed(self, button, modalias, pkg_name=None):
+    if self.ui_building:
+      return
+
+    pkg = None
+    try:
+      if pkg_name:
+        pkg = self.apt_cache[pkg_name]
+    except KeyError:
+      pass
+
+    if button.get_active():
+      if pkg in self.driver_changes:
+        self.driver_changes.remove(pkg)
+
+      if (pkg is not None 
+         and modalias in self.orig_selection
+         and button is not self.orig_selection[modalias]):
+        self.driver_changes.append(pkg)
+    else:
+      if pkg in self.driver_changes:
+        self.driver_changes.remove(pkg)
+
+      # for revert; to re-activate the original radio buttons.
+      if modalias not in self.orig_selection:
+        self.orig_selection[modalias] = button
+
+      if (pkg is not None
+         and pkg not in self.driver_changes
+         and pkg.is_installed):
+        self.driver_changes.append(pkg)
+
+    self.button_driver_revert.set_sensitive(bool(self.driver_changes))
+    self.button_driver_apply.set_sensitive(bool(self.driver_changes))
+
+  def gather_device_data(self, device):
+    '''Get various device data used to build the GUI.
+
+      return a tuple of (overall_status string, icon, drivers dict).
+      the drivers dict is using this form:
+        {"recommended/alternative": {pkg_name: {
+                                                  'selected': True/False
+                                                  'description': 'description'
+                                                  'builtin': True/False
+                                                }
+                                     }}
+         "manually_installed": {"manual": {'selected': True, 'description': description_string}}
+         "no_driver": {"no_driver": {'selected': True/False, 'description': description_string}}
+
+         Please note that either manually_installed and no_driver are set to None if not applicable
+         (no_driver isn't present if there are builtins)
+    '''
+
+    possible_overall_status = {
+      'recommended': (_("This device is using the recommended driver."), "recommended-driver"),
+      'alternative': (_("This device is using an alternative driver."), "other-driver"),
+      'manually_installed': (_("This device is using a manually-installed driver."), "other-driver"),
+      'no_driver': (_("This device is not working."), "disable-device")
+    }
+
+    returned_drivers = {'recommended': {}, 'alternative': {}, 'manually_installed': {}, 'no_driver': {}}
+    have_builtin = False
+    one_selected = False
+    try:
+      if device['manual_install']:
+        returned_drivers['manually_installed'] = {True: {'selected': True,
+                                                  'description': _("Continue using a manually installed driver")}}
+    except KeyError:
+      pass
+
+    for pkg_driver_name in device['drivers']:
+      current_driver = device['drivers'][pkg_driver_name]
+
+      # get general status
+      driver_status = 'alternative'
+      try:
+        if current_driver['recommended'] and current_driver['from_distro']:
+          driver_status = 'recommended'
+      except KeyError:
+        pass
+
+      builtin = False
+      try:
+        if current_driver['builtin']:
+          builtin = True
+          have_builtin = True
+      except KeyError:
+        pass
+
+      try:
+        pkg = self.apt_cache[pkg_driver_name]
+        installed = pkg.is_installed
+        description = _("Using {} from {}").format(pkg.candidate.summary, pkg.shortname)
+      except KeyError:
+        print("WARNING: a driver ({}) doesn't have any available package associated: {}".format(pkg_driver_name, current_driver))
+        continue
+
+      # gather driver description
+      if current_driver['free']:
+        licence = _("open source")
+      else:
+        licence = _("proprietary")
+
+      if driver_status == 'recommended':
+        base_string = _("{base_description} ({licence}, tested)")
+      else:
+        base_string = _("{base_description} ({licence})")
+      description = base_string.format(base_description=description, licence=licence)
+
+      selected = False
+      if not builtin and not returned_drivers['manually_installed']:
+        selected = installed
+        if installed:
+          selected = True
+          one_selected = True
+
+      returned_drivers[driver_status].setdefault(pkg_driver_name, {'selected': selected,
+                                                                   'description': description,
+                                                                   'builtin': builtin})
+
+    # adjust making the needed addition
+    if not have_builtin:
+      selected = False
+      if not one_selected:
+        selected = True
+      returned_drivers["no_driver"] = {True: {'selected': selected,
+                                              'description': _("Do not use the device")}}
+    else:
+      # we have a builtin and no selection: builtin is the selected one then
+      if not one_selected:
+        for section in ('recommended', 'alternative'):
+          for pkg_name in returned_drivers[section]:
+            if returned_drivers[section][pkg_name]['builtin']:
+              returned_drivers[section][pkg_name]['selected'] = True
+
+    # compute overall status
+    for section in returned_drivers:
+      for keys in returned_drivers[section]:
+        if returned_drivers[section][keys]['selected']:
+          (overall_status, icon) = possible_overall_status[section]
+
+    return (overall_status, icon, returned_drivers)
+
+
+  def show_drivers(self):
+    self.ui_building = True
+    self.dynamic_device_status = {}
+    for device in sorted(self.devices.keys()):
+      (overall_status, icon, drivers) = self.gather_device_data(self.devices[device])
+
+      driver_status = Gtk.Image()
+      driver_status.set_valign(Gtk.Align.START)
+      driver_status.set_halign(Gtk.Align.CENTER)
+      driver_status.set_from_icon_name(icon, Gtk.IconSize.MENU)
+      device_box = Gtk.Box(spacing=6, orientation=Gtk.Orientation.HORIZONTAL)
+      device_box.pack_start(driver_status, False, False, 6)
+      device_detail = Gtk.Box(spacing=6, orientation=Gtk.Orientation.VERTICAL)
+      device_box.pack_start(device_detail, True, True, 0)
+
+      widget = Gtk.Label("{}: {}".format(self.devices[device].get('vendor', _('Unknown')), self.devices[device].get('model', _('Unknown'))))
+      widget.set_halign(Gtk.Align.START)
+      device_detail.pack_start(widget, True, False, 0)
+      widget = Gtk.Label("<small>{}</small>".format(overall_status))
+      widget.set_halign(Gtk.Align.START)
+      widget.set_use_markup(True)
+      device_detail.pack_start(widget, True, False, 0)
+      self.dynamic_device_status[device] = (driver_status, widget)
+
+      option_group = None
+      # define the order of introspection
+      for section in ('recommended', 'alternative', 'manually_installed', 'no_driver'):
+        for driver in drivers[section]:
+          radio_button = Gtk.RadioButton.new_with_label(None, drivers[section][driver]['description'])
+          if option_group:
+            radio_button.join_group(option_group)
+          else:
+            option_group = radio_button
+          device_detail.pack_start(radio_button, True, False, 0)
+          radio_button.set_active(drivers[section][driver]['selected'])
+
+          if section == 'no_driver':
+            self.no_drv.append(radio_button)
+          if section in ('manually_install', 'no_driver') or ('builtin' in drivers[section][driver] and drivers[section][driver]['builtin']):
+            radio_button.connect("toggled", self.on_driver_selection_changed, device)
+          else:
+            radio_button.connect("toggled", self.on_driver_selection_changed, device, driver)
+          if drivers['manually_installed'] and section != 'manually_installed':
+            radio_button.set_sensitive(False)
+
+      self.box_driver_detail.pack_start(device_box, False, False, 6)
+
+    self.ui_building = False
+    self.box_driver_detail.show_all()
+    self.set_driver_action_status()
+
+  def update_label_and_icons_from_status(self):
+    '''Update the current label and icon, computing the new device status'''
+
+    for device in self.devices:
+      (overall_status, icon, drivers) = self.gather_device_data(self.devices[device])
+      (driver_status, widget) = self.dynamic_device_status[device]
+
+      driver_status.set_from_icon_name(icon, Gtk.IconSize.MENU)
+      widget.set_label("<small>{}</small>".format(overall_status))
+
+
+  def set_driver_action_status(self):
+    # Update the label in case we end up having some kind of proprietary driver in use.
+    if (os.path.exists('/var/run/reboot-required')):
+      self.label_driver_action.set_label(_("You need to restart the computer to complete the driver changes."))
+      self.button_driver_restart.set_visible(True)
+      self.window_main.set_urgency_hint(True)
+      return
+
+    self.nonfree_drivers = 0
+    for device in self.devices:
+      for pkg_name in self.devices[device]['drivers']:
+        pkg = self.apt_cache[pkg_name]
+        if not self.devices[device]['drivers'][pkg_name]['free'] and pkg.is_installed:
+          self.nonfree_drivers = self.nonfree_drivers + 1
+
+    if self.nonfree_drivers > 0:
+      self.label_driver_action.set_label(gettext.ngettext (
+                                                "%(count)d proprietary driver in use.",
+                                                "%(count)d proprietary drivers in use.", 
+                                                self.nonfree_drivers)
+                                                % { 'count': self.nonfree_drivers})
+    else:
+      self.label_driver_action.set_label(_("No proprietary drivers are in use."))
+
